import pandas as pd
import numpy as np
from scipy.stats import norm
from ramsmod.utils import convert_to_pd_series

__all__ = ['kaplan_meier_fit', 'turnbull_fit']


def kaplan_meier_fit(t, d, ci=None, alpha=0.05):
    """
    Produces a table from right-censored failure data containing data relating to
    the Kaplan-Meier estimate of the reliability function.
    :param t: Survival times of the observations in the failure data.
    :param d: Indicator variable values showing if observations were failures (value 1) or right-censored (value 0).
    :param ci: string with value 'gw' for Greenwood's and 'egw' for exponential
    Greenwood's confidence interval bounds.
    :param alpha: float giving level of significance for confidence interval (i.e.
    giving (1-alpha)*100% confidence level it contains true reliability). Default is 0.05.
    :return: Pandas DataFrame with columns 't', 'm', 'n' and 'R' containing
     the ordered failure times and corresponding number of observed failures,
     number at risk and Kaplan-Meier reliability estimates respectively.
   """
    t = convert_to_pd_series(t)
    d = convert_to_pd_series(d)

    if t.size != d.size:
        raise ValueError("times and observed must be equal size.")

    # Get the ordered failure times.
    only_failures = t[d == 1]
    failures_with_zero = only_failures.append(
        pd.Series([0]), ignore_index=True)
    unique_failures = pd.Series(failures_with_zero.unique())
    tf = unique_failures.sort_values(ignore_index=True)

    # Get the number of failures observed at the ordered failure times.
    mf = tf.apply(lambda x: (only_failures == x).sum())

    # Get the number right-censored prior in prior interval to each ordered failure time.
    only_censored = t[d == 0]
    total_q = tf.apply(lambda x: (only_censored < x).sum()).shift(periods=-1)
    qf = total_q.diff()
    qf.iloc[0] = total_q.iloc[0] # Set censored between t_(0) and t_(1).
    qf.iloc[-1] = (only_censored >= tf.iloc[-1]).sum() # Set censored after final ordered failure time.

    # Get the number at risk to fail at the ordered failure times.
    nf = tf.apply(lambda x: (t >= x).sum())

    # Calculate conditional survival probability
    # at the ordered failure times.
    surv_probs = 1 - (mf / nf)

    # Calculate Kaplan-Meier reliability estimates at the
    # ordered failure times.
    r = surv_probs.cumprod()

    # Create Pandas DataFrame with results.
    km_table = pd.DataFrame({'t': tf, 'm': mf, 'q':qf, 'n': nf,
                             'R': r})
    km_table.index.name = "f"

    # Add confidence intervals.
    if ci == 'gw':
        km_table['CI Lower'], km_table['CI Upper'] =  _greenwoods_ci(mf, nf, r, alpha)
    elif ci == 'egw':
        km_table['CI Lower'], km_table['CI Upper'] = _exp_greenwoods_ci(mf, nf, r, alpha)
    elif ci is not None:
        raise ValueError("Invalid ci value, must be 'gw' for Greenwood's or 'egw' for exponential Greenwood's"
                         " confidence interval.")

    return km_table


def _get_greenwoods_formula_cumsums(mf, nf):
    """
    Gets the value of the sum term from Greenwood's confidence
    interval at each ordered failure time.
    """
    sum_terms = (mf / (nf * (nf - mf)))
    sum_terms = sum_terms.replace([np.inf], 0) # No interval where R(t)=0.
    cumulative_sums = np.cumsum(sum_terms)

    return cumulative_sums

def _greenwoods_ci(mf, nf, r, alpha=0.05):
    """
    Gets the confidence interval given by Greenwood's formula.
    """
    if alpha <= 0 or alpha >= 1:
        raise ValueError("alpha must be between 0 and 1 exclusive.")

    cumulative_sums = _get_greenwoods_formula_cumsums(mf, nf)
    # Compute the variance values given by Greenwood's formula at
    # each ordered failure time.
    var = np.power(r, 2) * cumulative_sums
    # Compute the standard deviations at each ordered failure time.
    sd = np.sqrt(var)
    z = norm.ppf(1 - alpha / 2)  # Z-score corresponding to the alpha / 2.
    # Compute the lower and upper bounds of the confidence interval.
    ci_lb = r - z * sd
    ci_ub = r + z * sd

    return ci_lb, ci_ub


def _exp_greenwoods_ci(mf, nf, r, alpha=0.05):
    """
    Gets the exponential Greenwood's confidence interval.
    """
    if alpha <= 0 or alpha >= 1:
        raise ValueError("alpha must be between 0 and 1 exclusive.")

    # Don't calculate confidence interval for time 0 where R(t)=1.
    mf = mf[1:]
    nf = nf[1:]
    r = r[1:]

    with np.errstate(divide='ignore'):
        cumulative_sums = _get_greenwoods_formula_cumsums(mf, nf)
        # Compute the V value in exponential Greenwood's confidence interval.
        v = (1 / np.power(np.log(r), 2)) * cumulative_sums
        z = norm.ppf(1 - alpha / 2)  # Z-score corresponding to the alpha / 2.
        # Compute the c+- values in exponential Greenwood's confidence interval.
        c_plus = np.log(-np.log(r)) + z * np.sqrt(v)
        c_minus = np.log(-np.log(r)) - z * np.sqrt(v)
        # Compute the lower and upper bounds of the confidence interval.
        ci_lb = np.exp(-np.exp(c_plus))
        ci_ub = np.exp(-np.exp(c_minus))

    # Pre-pend confidence interval for time 0 of (1.0, 1.0).
    ci_lb = np.concatenate((pd.Series([1.0]), ci_lb))
    ci_ub = np.concatenate((pd.Series([1.0]), ci_ub))

    return ci_lb, ci_ub


def turnbull_fit(tmin, tmax, tol=0.001):
    """
    Computes Turnbull estimates of reliability from interval censored
    failure data.

    Parameters
    ----------
    tmin: Lower bounds of failure times for observations.
    tmax: Upper bounds of failure times for observations.
    tol: Tolerance for the iterative procedure, convergence terminates when
     maximum difference from previous reliability estimate at any time is less than
     tolerance.

    Returns
    -------
    Pandas DataFrame with column 't' containing the interval end point time values
     and 'R' containing the corresponding Turnbull reliability estimates.
    """
    tmin = convert_to_pd_series(tmin)
    tmax = convert_to_pd_series(tmax)

    if tmin.size != tmax.size:
        raise ValueError("t_min and t_max must be equal size.")

    t = np.sort(pd.Series([0]).append(tmin).append(tmax).unique())

    # Initial reliability function as equal reduction at each time grid point.
    reliabilities = np.linspace(1.0, 0, len(t))

    # Form alphas matrix describing if each observation (matrix rows)
    # could fail (value 1) or not (value 0) in each interval (matrix columns).
    n = len(tmin)
    m = len(t) - 1
    alphas = np.empty((n, m), dtype=np.bool)
    for i in range(n):
        t_min_i = tmin.iloc[i]
        t_max_i = tmax.iloc[i]
        for j in range(m):
            grid_t_min_j = t[j]
            grid_t_max_j = t[j + 1]
            if t_min_i != t_max_i:
                alphas[(i, j)] = t_min_i < grid_t_max_j and t_max_i > grid_t_min_j
            else:  # Exact failure observation.
                alphas[(i, j)] = t_min_i <= grid_t_max_j and t_max_i > grid_t_min_j

    while True:
        # Compute estimated failures in each interval.
        p = -np.diff(reliabilities)  # Interval failure probabilities.
        p_alphas = alphas * p
        d = ((p_alphas.T / p_alphas.sum(axis=1)).T.sum(axis=0))
        # Compute number at risk in each interval.
        y = np.cumsum(d[::-1])[::-1]
        updated_reliabilities = np.insert(np.cumprod(1 - (d / y)), 0, 1)

        difference = np.max(np.abs(updated_reliabilities - reliabilities))
        reliabilities = updated_reliabilities
        if difference <= tol:
            break

    turnbull_table = pd.DataFrame({'t': t, 'R': reliabilities}, index=range(1, m+2))

    return turnbull_table





import numpy as np
import pandas as pd


def generate_right_censored(ttf_rv, sojourn_rv, size, decimals=0, sort=True):
    """
    Generates a table of right-censored failure data with random censoring.
    :param ttf_rv: A continuous distribution from scipy.stats for the times to failure,
    e.g. weibull_min(1.4, scale=100).
    :param sojourn_rv: A continuous distribution from scipy.stats for the maximum
    observation time, e.g. uniform(0, 100). Right-censoring occurs for an
    observation if this time is less than the time to failure.
    :param size: Number of observations to generate.
    :param decimals: Number of decimals to round the times.
    :param sort: If True then observations are sorted in order of increasing observed
     survival time.
    :return: Pandas DataFrame where the index gives the observation number and
     column 't' gives the survival time (failure or censorship) and column 'd' gives the value of an indicator
     variable that has value 1 if failure was observed and value 0 if
     failure was right-censored.
    """
    ttf = ttf_rv.rvs(size=size)
    censor_times = sojourn_rv.rvs(size=size)
    t = np.minimum(ttf, censor_times).round(decimals)
    d = (ttf <= censor_times).astype(int)

    ds = pd.DataFrame({'t': t, 'd': d})
    if sort:
        ds = ds.sort_values(by=['t'], ignore_index=True)
    ds.index.name = "Observation #"
    ds.index += 1

    return ds


def generate_interval_censored(ttf_rv, inspection_interval, size, entry_rv=None,
                               sojourn_rv=None):
    """
    Generates a table of interval-censored failure data where failures are only revealed
    at inspections performed periodically at the end of each fixed duration interval
     from the as-new condition.
    :param ttf_rv: A distribution from scipy.stats from which the
     times to failure are sampled, e.g. weibull_min(1.4, scale=100).
    :param inspection_interval: Duration of time between inspections.
    :param size: Number of observations to generate.
    :param entry_rv: A discrete distribution from scipy.stats
    describing the number of periods that complete before an item enters observation
    (e.g. 0 if enters at time 0, 1 if enters after 1 period is completed and misses
    the first inspection etc.)
    :param sojourn_rv: A discrete distribution from scipy.stats describing how many complete periods
    an item remains under observation after entry (e.g. 2 if it will be inspected twice before
    failure is right-censored if not yet revealed).
    :return Pandas DataFrame where the index gives the observation number and
     columns 'tmin' and 'tmax' contain the corresponding lower
     and upper bounds of the interval in which failure occurred (upper bound
     is set to np.infty for right-censored observations).
    """

    # Sample failure time, entry period and sojourn periods.
    ttf = ttf_rv.rvs(size=size)
    if entry_rv :
        entry_at_start_period = entry_rv.rvs(size=size)
    else:
        entry_at_start_period = np.zeros(size)

    if sojourn_rv:
        sojourn_periods = sojourn_rv.rvs(size=size)
    else:
        sojourn_periods = np.ones(size) * np.inf

    # Compute the period at which an observation exits and fails.
    # Assume first period is period 0, second is period 1 etc.
    exit_at_start_period = entry_at_start_period + sojourn_periods
    failed_in_period = np.where(np.mod(ttf, inspection_interval) != 0, np.floor(ttf / inspection_interval),
                                (ttf / inspection_interval)-1)  # Period includes end point not start point.

    # Calculate failure interval bounds.
    tmin = np.where(failed_in_period <= entry_at_start_period, 0, np.minimum(failed_in_period,
                                                                 exit_at_start_period)) * inspection_interval

    tmax = np.where(failed_in_period < exit_at_start_period, np.maximum(failed_in_period, entry_at_start_period) + 1,
                     np.infty) * inspection_interval

    ds = pd.DataFrame({'tmin':tmin, 'tmax': tmax}, index=range(1,size+1))
    ds.index.name = "Observation #"

    return ds



import pandas as pd
from pathlib import Path
import os

__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))


def load_right_censored_data():
    """ Returns a set of right-censored failure data comprising of 26 observations as a
    Pandas Dataframe with index of the observation number, column 't' giving
    the survival time and column 'd' an indicator variable where 1 means
    the an exact failure time was observed and 0 indicating failure was
    right-censored.  """
    pathname = os.path.join(__location__, 'right-censored.csv')
    return pd.read_csv(pathname, index_col='Observation #')


def load_right_censored_data2():
    """ Returns a set of right-censored failure data comprising of 22 observations as a
    Pandas Dataframe with index of the observation number, column 't' giving
    the survival time and column 'd' an indicator variable where 1 means
    the an exact failure time was observed and 0 indicating failure was
    right-censored.  """
    pathname = os.path.join(__location__, 'right-censored-2.csv')
    return pd.read_csv(pathname, index_col='Observation #')


def load_interval_censored_data():
    """ Returns a set of interval-censored failure data comprising of 26 observations as a
    Pandas Dataframe with index of the observation number, column 'tmin' giving
    the lower bound of the failure time interval and column 'tmax' giving the upper bound of the
     failure time interval."""
    pathname = os.path.join(__location__, 'interval-censored.csv')
    return pd.read_csv(pathname, index_col='Observation #')


def load_interval_censored_data2():
    """ Returns a set of interval-censored failure data comprising of 24 observations as a
    Pandas Dataframe with index of the observation number, column 'tmin' giving
    the lower bound of the failure time interval and column 'tmax' giving the upper bound of the
     failure time interval."""
    pathname = os.path.join(__location__, 'interval-censored-2.csv')
    return pd.read_csv(pathname, index_col='Observation #')


def load_exponential_data():
    """ Returns a set of right-censored failure data comprising of 100 observations as a
    Pandas Dataframe with index of the observation number, column 't' giving
    the survival time and column 'd' an indicator variable where 1 means
    the an exact failure time was observed and 0 indicating failure was
    right-censored. This data was generated from an exponential distribution
    with a scale parameter value of 300 and uniform random right-censoring on the
    interval (0, 400). In total, 48 failure times were observed exactly and 52
    were right-censored."""
    pathname = os.path.join(__location__, 'exponential_data.csv')
    return pd.read_csv(pathname, index_col='Observation #')


def load_weibull_data():
    """ Returns a set of right-censored failure data comprising of 100 observations as a
    Pandas Dataframe with index of the observation number, column 't' giving
    the survival time and column 'd' an indicator variable where 1 means
    the an exact failure time was observed and 0 indicating failure was
    right-censored. This data was generated from a Weibull distribution
    with a shape parameter of 2 and a scale parameter value of 250, with
    uniform random right-censoring on the interval (0, 400). In total, 52 failure
    times were observed exactly and 48 were right-censored."""
    pathname = os.path.join(__location__, 'weibull_data.csv')
    return pd.read_csv(pathname, index_col='Observation #')


def load_lognormal_data():
    """ Returns a set of right-censored failure data comprising of 100 observations as a
    Pandas Dataframe with index of the observation number, column 't' giving
    the survival time and column 'd' an indicator variable where 1 means
    the an exact failure time was observed and 0 indicating failure was
    right-censored. This data was generated from an log-normal distribution
    with a shape parameter of 0.5 and a scale parameter value of 170, with
    uniform random right-censoring on the interval (0,400). In total, 51 failure
    times were observed exactly and 49 were right-censored."""
    pathname = os.path.join(__location__, 'lognormal_data.csv')
    return pd.read_csv(pathname, index_col='Observation #')


import numpy as np
import pandas as pd


def convert_to_pd_series(l):
    """Convert l to a Pandas series if it's a list or np.ndarray."""
    if isinstance(l, (list, np.ndarray)):
        l = pd.Series(l)
    return l


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.ticker import MaxNLocator
from scipy.stats import norm, expon, lognorm, weibull_min, chi2
from ramsmod.utils import convert_to_pd_series

__all__ = ['plot_right_censored', 'plot_interval_censored', 'plot_np_reliability',
           'plot_exponential_prob_plot', 'plot_weibull_prob_plot', 'plot_lognormal_prob_plot',
           'plot_exponential_relative_likelihoods', 'plot_weibull_relative_likelihoods',
           'plot_lognormal_relative_likelihoods']


def plot_right_censored(t, d, ax=None, show_legend=True):
    """
    Returns a plot of observations from right-censored failure data.
    :param t: Survival times for each observation.
    :param d: Indicator variable value for each observation, where
    value 1 indicates exact failure observed and 0 indicates failure was right-censored.
    :param ax: Matplotlib axes on which to plot, if None then one will be created.
    :param show_legend: Boolean that is True if legend should be added to axes and
    False otherwise.

    :return: Matplotlib axes containing the plot.
    """
    t = convert_to_pd_series(t)
    d = convert_to_pd_series(d)

    if ax is None:
        ax = plt.gca()

    # Add the observations to the axes.
    for i in range(len(t)):
        if d.iloc[i]:
            # Failure observation.
            ax.scatter(t.iloc[i], i+1, color='b', marker='o')
        else:
            # Right-censored observation.
            ax.scatter(t.iloc[i], i+1, color='b', marker='>')

    ax.yaxis.set_major_locator(MaxNLocator(integer=True))
    ax.set_ylabel('Observation #')

    # Add legend to axes.
    if show_legend:
        legend_elements = [
            Line2D([0], [0], color='w', markeredgecolor='b',
                   markerfacecolor='b', marker='>',
                   label='Right-censored'),
            Line2D([0], [0], color='w',markeredgecolor='b',
                   markerfacecolor='b', marker='o',
                   label='Exact')]
        ax.legend(handles=legend_elements)

    return ax


def plot_interval_censored(tmin, tmax, ax=None, show_legend=True):
    """
    Returns a plot of observations from interval-censored failure data.
    :param tmin: The exclusive lower bounds of the failure time intervals.
    :param tmax: The inclusive upper bound of the failure time intervals
    (use np.infty for right-censored observations).
    :param ax: Matplotlib axes on which to plot, if None then one will be created.
    :param show_legend: Boolean that is True if legend should be added to axes and
    False otherwise.
    :return: Matplotlib axes containing the plot.
    """
    tmin = convert_to_pd_series(tmin)
    tmax = convert_to_pd_series(tmax)

    if ax is None:
        fig = plt.figure()  # Create plot figure.
        ax = fig.add_subplot()  # Create the axes to plot on.

    # Add the observations to the axes.
    for i in range(len(tmin)):
        if tmin.iloc[i] == tmax.iloc[i]:
            ax.scatter(tmin.iloc[i], i + 1, color='b', marker='o')
        elif tmax.iloc[i] == np.infty:
            ax.scatter(tmin.iloc[i], i + 1, color='b', marker='>')
        else:
            ax.hlines(i + 1, tmin.iloc[i], tmax.iloc[i], color='b')

    ax.yaxis.set_major_locator(MaxNLocator(integer=True))
    ax.set_ylabel('Observation #')
    ax.set_xlabel('Time')
    ax.set_xlim(0)
    ax.set_ylim(0)

    # Add legend to axes.
    if show_legend:
        legend_elements = [Line2D([0], [0], color='b',
                                  label='Interval-censored'),
                           Line2D([0], [0], color='w', markeredgecolor='b',
                                  markerfacecolor='b', marker='>',
                                  label='Right-censored'),
                           Line2D([0], [0], color='w', markeredgecolor='b',
                                  markerfacecolor='b', marker='o',
                                  label='Exact')]
        ax.legend(handles=legend_elements, loc='upper right')

    return ax


def plot_np_reliability(t, r, ci_lb=None, ci_ub=None, ax=None, linestyle='-',
                        color='blue', label=None):
    """
    Produces a step plot of reliability estimates against time.
    :param t: The time values (Panda's series, list etc.).
    :param r: The reliability estimates (Panda's series, list etc.) corresponding
    to the time values.
    :param ci_lb: The reliability estimates corresponding to the time values
    at the lower bound of the confidence interval.
    :param ci_ub: The reliability estimates corresponding to the time values
    at the upper bound of the confidence interval.
    :param ax: The Matplotlib axes on which to draw the plot.
    :param linestyle: The Matplotlib linestyle option for the plot.
    :param color: The Matplotlib color style option for the plot.
    :param label: The label for the plot.
    :return: The Matplotlib axes containing the plot.
    """
    if ax is None:
        ax = plt.gca()

    # Plot r against t as step function.
    ax.step(t, r, where='post', color=color,
            label=label, linestyle=linestyle)

    # Add confidence intervals if provided.
    if ci_lb is not None:
        ax.step(t, ci_lb, where='post', color=color, alpha=0.5,
                linestyle=linestyle)
    if ci_ub is not None:
        ax.step(t, ci_ub, where='post', color=color, alpha=0.5,
                linestyle=linestyle)

    # Add axis-labels.
    ax.set_xlabel(r'$t$')
    ax.set_ylabel(r'$\widehat R(t)$')

    ax.set_xlim(0)

    return ax


def plot_exponential_prob_plot(t, r, ci_lb=None, ci_ub=None, ax=None, marker='o',
                        color='blue', label=None):
    """
    Produces a linearised probability plot for the exponential distribution.
    :param t: The time values.
    :param r: The reliability estimates corresponding to the time values.
    :param ci_lb: The reliability estimates corresponding to the time values
    at the lower bound of the confidence interval.
    :param ci_ub: The reliability estimates corresponding to the time values
    at the upper bound of the confidence interval.
    :param ax: The Matplotlib axes on which to draw the plot.
    :param marker: The Matplotlib marker style option for the plot.
    :param color: The Matplotlib color style option for the plot.
    :param label: The label for the plot.
    :return: The Matplotlib axes containing the plot.
    """
    if ax is None:
        ax = plt.gca()

    y = -np.log(r)
    ax.scatter(t, y, color=color,
            label=label, marker=marker)
    if ci_lb is not None:
        y_lb = -np.log(ci_lb)
        ax.plot(t, y_lb, color=color, alpha=0.5,
                   label=label, linestyle='--')
    if ci_ub is not None:
        y_ub = -np.log(ci_ub)
        ax.plot(t, y_ub, color=color, alpha=0.5,
                  label=label, linestyle='--')

    # Add axis-labels.
    ax.set_xlabel(r'$t$', fontsize=12)
    ax.set_ylabel(r'$-\ln(\widehat R(t))$', fontsize=12)

    return ax


def plot_weibull_prob_plot(t, r, ci_lb=None, ci_ub=None, ax=None, marker='o',
                                   color='blue', label=None):
    """
    Produces a linearised probability plot for the Weibull distribution.
    :param t: The time values.
    :param r: The reliability estimates corresponding to the time values.
    :param ci_lb: The reliability estimates corresponding to the time values
    at the lower bound of the confidence interval.
    :param ci_ub: The reliability estimates corresponding to the time values
    at the upper bound of the confidence interval.
    :param ax: The Matplotlib axes on which to draw the plot.
    :param marker: The Matplotlib marker style option for the plot.
    :param color: The Matplotlib color style option for the plot.
    :param label: The label for the plot.
    :return: The Matplotlib axes containing the plot.
    """
    if ax is None:
        ax = plt.gca()

    y = np.log(-np.log(r))
    x = np.log(t)
    ax.scatter(x, y, color=color,
               label=label, marker=marker)
    if ci_lb is not None:
        y_lb = np.log(-np.log(ci_lb))
        ax.plot(x, y_lb, color=color, alpha=0.5,
                   label=label, linestyle='--')
    if ci_ub is not None:
        y_ub = np.log(-np.log(ci_ub))
        ax.plot(x, y_ub, color=color, alpha=0.5,
                   label=label, linestyle='--')

    # Add axis-labels.
    ax.set_xlabel(r'$\ln(t)$', fontsize=12)
    ax.set_ylabel(r'$\ln(-\ln(\widehat R(t)))$', fontsize=12)

    return ax


def plot_lognormal_prob_plot(t, r, ci_lb=None, ci_ub=None, ax=None, marker='o',
                           color='blue', label=None):
    """
    Produces a linearised probability plot for the log-normal distribution.
    :param t: The time values.
    :param r: The reliability estimates corresponding to the time values.
    :param ci_lb: The reliability estimates corresponding to the time values
    at the lower bound of the confidence interval.
    :param ci_ub: The reliability estimates corresponding to the time values
    at the upper bound of the confidence interval.
    :param ax: The Matplotlib axes on which to draw the plot.
    :param marker: The Matplotlib marker style option for the plot.
    :param color: The Matplotlib color style option for the plot.
    :param label: The label for the plot.
    :return: The Matplotlib axes containing the plot.
    """
    if ax is None:
        ax = plt.gca()

    y = norm.ppf(1-r)
    x = np.log(t)
    ax.scatter(x, y, color=color,
               label=label, marker=marker)
    if ci_lb is not None:
        y_lb = norm.ppf(1-ci_lb)
        ax.plot(x, y_lb, color=color, alpha=0.5,
                   label=label, linestyle='--')
    if ci_ub is not None:
        y_ub = norm.ppf(1-ci_ub)
        ax.plot(x, y_ub, color=color, alpha=0.5,
                   label=label, linestyle='--')

    # Add axis-labels.
    ax.set_xlabel(r'$\ln(t)$', fontsize=12)
    ax.set_ylabel(r'$\sigma^{-1}(1-\widehat R(t))$', fontsize=12)

    return ax


def _get_exponential_relative_log_likelihoods(tmin, tmax, scales):

    total_lls = np.zeros(scales.shape)

    interval_t = tmin != tmax
    exact_t = tmin == tmax

    for i in range(len(scales)):
        # Compute likelihood.
        scale = scales[i]

        interval_log_likelihoods = np.log(expon.cdf(tmax[interval_t], scale=scale) -
                                 expon.cdf(tmin[interval_t], scale=scale))
        exact_log_likelihoods = np.log(expon.pdf(tmin[exact_t], scale=scale))
        total_log_likelihood = interval_log_likelihoods.sum() + exact_log_likelihoods.sum()
        total_lls[i] = total_log_likelihood

    max_ll = max(total_lls)
    relative_likelihoods = np.exp(total_lls - max_ll)

    return relative_likelihoods


def _get_total_log_likelihood(ttf_rv, tmin, tmax):
    interval_t = tmin != tmax
    exact_t = tmin == tmax

    interval_log_likelihoods = np.log(ttf_rv.cdf(tmax[interval_t]) - ttf_rv.cdf(tmin[interval_t]))
    exact_log_likelihoods = np.log(ttf_rv.pdf(tmin[exact_t]))
    total_log_likelihood = interval_log_likelihoods.sum() + exact_log_likelihoods.sum()

    return total_log_likelihood


def _get_shape_scale_relative_likelihoods(tmin, tmax, shapes, scales, dist_fun):

    total_log_likelihoods = np.zeros((shapes.shape[0], scales.shape[0]))

    for i in range(len(shapes)):
        for j in range(len(scales)):
            # Compute likelihood.
            shape = shapes[i]
            scale = scales[j]

            ttf_rv = dist_fun(shape, scale=scale)

            # Compute total log-likelihood of observations in the failure data.
            total_log_likelihood = _get_total_log_likelihood(ttf_rv, tmin, tmax)

            # Compute the total log-likelihood and store in array.
            total_log_likelihoods[i, j] = total_log_likelihood

    # Compute maximum log-likelihood.
    max_log_likelihood = np.max(total_log_likelihoods)

    # Compute relative likelihoods.
    relative_likelihoods = np.exp(total_log_likelihoods - max_log_likelihood)

    return relative_likelihoods


def plot_exponential_relative_likelihoods(tmin, tmax, min_scale, max_scale, step_scale, ci_alpha=0.05,
                                          ax=None, linestyle='-', color='blue', label=None):
    """
    Plots the relative likelihoods corresponding to scale parameter value of the exponential
    distribution for a set of interval censored failure data, including a confidence interval
    for the true scale parameter value.
    :param tmin: The lower bounds for the failure time observations.
    :param tmax: The upper bounds for the failure time observations.
    :param min_scale: The minimum value for the scale parameter included on the plot.
    :param max_scale: The maximum value for the scale parameter included on the plot.
    :param step_scale: The step (increment) between plotted scale parameter values.
    :param ci_alpha: The significance value for the confidence interval (i.e.
    producing (1-ci_alpha)*100% confidence interval).
    :param ax: The Matplotlib axes on which to draw the plot, new axes created if None.
    :param linestyle: The Matplotlib linestyle option for the plot.
    :param color: The Matplotlib color option for the plot.
    :param label: The label for the plot line.
    :return: The Matplotlib axes containing the plot.
    """
    if ax is None:
        ax = plt.gca()

    scales = np.arange(min_scale, max_scale, step_scale)
    relative_likelihoods = _get_exponential_relative_log_likelihoods(tmin, tmax, scales)
    ax.plot(scales, relative_likelihoods, linestyle=linestyle, color=color, label=label)

    # Confidence interval.
    ci_relative_likelihood = np.exp(-chi2.ppf(1 - ci_alpha, df=1) / 2)

    ci_region = relative_likelihoods >= ci_relative_likelihood
    ci_region_likelihoods = relative_likelihoods[ci_region]
    ci_region_scales = scales[ci_region]

    ax.fill_between(ci_region_scales, ci_region_likelihoods, color=color, alpha=0.5)
    ax.set_xlabel(r'$\Theta$')
    ax.set_ylabel(r'$R(\Theta)$')
    ax.set_ylim(0)

    return ax


def _plot_shape_scale_confidence_regions(tmin, tmax, min_shape, max_shape, step_shape,
                                         min_scale, max_scale, step_scale, dist_fun,
                                         label_as_confidence, ax=None):

    if ax is None:
        ax = plt.gca()

    shapes = np.arange(min_shape, max_shape, step_shape)
    scales = np.arange(min_scale, max_scale, step_scale)
    relative_likelihoods = _get_shape_scale_relative_likelihoods(tmin, tmax, shapes, scales, dist_fun)

    cs = ax.contour(scales, shapes, relative_likelihoods)
    levels = cs.levels

    if label_as_confidence:
        def format_confidence_levels(level):
            value = 100*chi2.cdf(-2 * np.log(level), df=2).round(2)
            return f"{value:.2f}%"

        ax.clabel(cs, levels, fmt=format_confidence_levels)
        ax.set_title('Confidence regions')
    else:
        ax.clabel(cs, levels)
        ax.set_title('Relative likelihood regions')

    return ax


def plot_weibull_relative_likelihoods(tmin, tmax, min_shape, max_shape, step_shape, min_scale,
                                      max_scale, step_scale, label_as_confidence=False, ax=None):
    """
    Plots (as a contour plot) the relative likelihoods corresponding to parameter values
    of the Weibull distribution for a set of interval censored failure data.
    :param tmin: The lower bounds for the failure time observations.
    :param tmax: The upper bounds for the failure time observations.
    :param min_shape: The minimum value for the shape parameter included on the plot.
    :param max_shape: The maximum value for the shape parameter included on the plot.
    :param step_shape: The step (increment) between plotted shape parameter values.
    :param min_scale: The minimum value for the scale parameter included on the plot.
    :param max_scale: The maximum value for the scale parameter included on the plot.
    :param step_scale: The step (increment) between plotted scale parameter values.
    :param label_as_confidence: If True then the confidence level contours are plotted
    instead of relative likelihoods.
    :param ax: The Matplotlib axes on which to draw the plot, if None then new axes created.
    :return: The Matplotlib axes containing the plot.
    """
    dist_fun = weibull_min
    ax = _plot_shape_scale_confidence_regions(tmin, tmax, min_shape, max_shape, step_shape,
                                              min_scale, max_scale, step_scale, dist_fun,
                                              label_as_confidence, ax=ax)
    ax.set_xlabel(r'$\eta$')
    ax.set_ylabel(r'$\beta$')

    return ax


def plot_lognormal_relative_likelihoods(tmin, tmax, min_shape, max_shape, step_shape,
                                        min_scale, max_scale, step_scale, label_as_confidence=False,
                                        ax=None):
    """
    Plots (as a contour plot) the relative likelihoods corresponding to parameter values
    of the log-normal distribution for a set of interval censored failure data.
    :param tmin: The lower bounds for the failure time observations.
    :param tmax: The upper bounds for the failure time observations.
    :param min_shape: The minimum value for the shape parameter included on the plot.
    :param max_shape: The maximum value for the shape parameter included on the plot.
    :param step_shape: The step (increment) between plotted shape parameter values.
    :param min_scale: The minimum value for the scale parameter included on the plot.
    :param max_scale: The maximum value for the scale parameter included on the plot.
    :param step_scale: The step (increment) between plotted scale parameter values.
    :param label_as_confidence: If True then the confidence level contours are plotted
    instead of relative likelihoods.
    :param ax: The Matplotlib axes on which to draw the plot, if None then new axes created.
    :return: The Matplotlib axes containing the plot.
    """
    dist_fun = lognorm
    ax = _plot_shape_scale_confidence_regions(tmin, tmax, min_shape, max_shape, step_shape,
                                              min_scale, max_scale, step_scale, dist_fun,
                                              label_as_confidence, ax=ax)
    plt.xlabel(r'$m$')
    plt.ylabel(r'$\sigma$')

    return ax


from setuptools import setup, find_packages

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

package_data = {"ramsmod": ["datasets/*"]}

setup(
  name = 'ramsmod',
  packages = find_packages(exclude=['*.tests',]),
  package_data = package_data,
  version = '0.6.2',      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'Python 3 library for reliability data analysis',   # Give a short description about your library
  author = 'Sean Reed',                   # Type in your name
  author_email = 'sean@sean-reed.com',      # Type in your E-Mail
  url = 'https://github.com/user/sean-reed',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/sean-reed/ramsmod/archive/v0.6.2.tar.gz',
  install_requires= requirements,
  classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
    ],
)


import numpy as np
import pandas as pd
from math import pow
from scipy.stats import chi2, norm
from ramsmod.utils import convert_to_pd_series

__all__ = ['log_rank_test', 'mantel_test']


def log_rank_test(t1, d1, t2, d2):
    """
    Performs a log-rank test to evaluate the null hypothesis that
    two groups have the same reliability from right-censored failure data.
    :param t1: Survival times for the observations in the failure data
    for group 1.
    :param d1: Indicator variable values showing if observations
    were failures (value 1) or right-censored (value 0) for group 1.
    :param t2: Survival times of the observations in the failure data
    for group 2.
    :param d2: Indicator variable values showing if observations
    were failures (value 1) or right-censored (value 0) for group 2.
    :return:  A tuple containing a Pandas DataFrame with a table of results from
    the calculations used to perform the test, the log-rank test statistic, the
    estimated variance of the statistic distribution and the calculated P-value for the test.
    """

    # Convert inputs to pd.Series if not already.
    t1 = convert_to_pd_series(t1)
    d1 = convert_to_pd_series(d1)
    t2 = convert_to_pd_series(t2)
    d2 = convert_to_pd_series(d2)

    t = pd.concat([t1, t2])
    d = pd.concat([d1, d2])
    # Ordered failure times.
    tf = pd.Series(t[d == 1].unique()).sort_values(ignore_index=True)
    # Observed failures.
    m1 = tf.apply(lambda x: sum(t1[d1 == 1] == x))
    m2 = tf.apply(lambda  x: sum(t2[d2 == 1] == x))
    # Number at risk.
    n1 = tf.apply(lambda x: sum(t1 >= x))
    n2 = tf.apply(lambda x: sum(t2 >= x))
    # Expected failures under null hypothesis.
    e1 = n1 / (n1 + n2) * (m1 + m2)
    e2 = n2 / (n1 + n2) * (m1 + m2)

    table = pd.DataFrame({'tf': tf, 'm1f': m1, 'm2f': m2, 'n1f': n1, 'n2f': n2,
                          'e1f': e1, 'e2f': e2})

    # Calculate log-rank statistic.
    num = (n1 * n2 * (m1 + m2) * (n1 + n2 - m1 - m2))
    den = (n1 + n2).pow(2) * (n1 + n2 - 1)
    var = sum((num / den).replace([np.nan], 0))
    log_rank_stat = pow(sum(m1) - sum(e1), 2) / var
    p = chi2(1).sf(log_rank_stat)

    return table, log_rank_stat, var, p


def mantel_test(t_min_1, t_max_1, t_min_2, t_max_2):
    """
    Performs a Mantel test to evaluate the null hypothesis that
    two groups have the same reliability from interval-censored failure data.
    :param t_min_1: Exclusive lower bounds of the failure intervals
    for the observations from the group 1 failure data.
    :param t_max_1: Inclusive upper bounds of the failure intervals
    for the observations from the group 1 failure data.
   :param t_min_2: Exclusive lower bounds of the failure intervals
    for the observations from the group 2 failure data.
    :param t_max_2: Inclusive upper bounds of the failure intervals
    for the observations from the group 2 failure data.
    :return: A tuple containing a Pandas DataFrame with a table containing results from
    calculations used to perform the test, the Mantel test statistic, the estimated
    variance in the test statistic and the calculated P-value for the test.
    """
    # Convert inputs to pd.Series if not already.
    t_min_1 = convert_to_pd_series(t_min_1)
    t_max_1 = convert_to_pd_series(t_max_1)
    t_min_2 = convert_to_pd_series(t_min_2)
    t_max_2 = convert_to_pd_series(t_max_2)

    t_min = pd.concat([t_min_1, t_min_2], ignore_index=True)
    t_max = pd.concat([t_max_1, t_max_2], ignore_index=True)

    n_1 = t_min_1.size
    n_2 = t_min_2.size
    n = n_1 + n_2

    later = np.zeros(n)
    earlier = np.zeros(n)
    for i in range(n):
        later[i] = sum(t_min[i] >= t_max)
        earlier[i] = sum(t_max[i] <= t_min)

    v = later - earlier

    table = pd.DataFrame({'t_min': t_min, 't_max': t_max, 'later': later,
                          'earlier': earlier, 'v': v}, index=range(1,n+1))
    table.index.name = "Observation #"

    var = n_1 * n_2 * sum(np.power(v, 2)) / ((n_1 + n_2) * (n_1 + n_2 - 1))
    sd = np.sqrt(var)
    w = sum(v[:n_1])
    p = norm.sf(abs(w), scale=sd)*2

    return table, w, var, p







